/*
 * Copyright 2016 Google Inc. All Rights Reserved.
 *
 * This code is confidential and provided to Android partners under NDA. You may modify, compile and
 * distribute the compiled code with your management software to assist with the provisioning and
 * management of Android devices. Other uses require written permission.
 */
package com.google.android.apps.work.dpcsupport;

import static com.google.android.apps.work.dpcsupport.PlayServicesAndStoreUtils.PLAY_SERVICES_MIN_VERSION;

import android.Manifest;
import android.app.ActivityManager;
import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.os.Build;
import android.os.Handler;
import android.provider.Settings;
import android.support.annotation.VisibleForTesting;
import android.support.annotation.WorkerThread;
import android.util.Log;
import com.google.android.apps.work.dpcsupport.PlayStoreDownloader.DownloadCompleteListener;
import com.google.android.apps.work.dpcsupport.WorkingEnvironmentCallback.Error;
import com.google.android.gms.common.GoogleApiAvailability;
import java.io.InputStream;
import java.util.Collections;

/**
 * Perform all necessary actions to add support for Android for Work Account provisioning.
 *
 * @see AndroidForWorkAccountSupport#ensureWorkingEnvironment
 */
class EnvironmentPreparer {

  private static final String[] REQUIRED_PERMISSIONS = {
      "android.permission.DOWNLOAD_WITHOUT_NOTIFICATION",
      "android.permission.GET_ACCOUNTS",
      "android.permission.MANAGE_ACCOUNTS",
      "android.permission.WRITE_SYNC_SETTINGS",
      "com.google.android.providers.gsf.permission.READ_GSERVICES"
  };

  // Progress Checkpoints.
  static final float PROGRESS_PLAY_STORE_CHECK_UPDATE = 0.05f;
  static final float PROGRESS_PLAY_STORE_DOWNLOADED = 0.25f;
  static final float PROGRESS_PLAY_STORE_UPDATE_FINISHED = 0.4f;
  static final float PROGRESS_PLAY_SERVICES_CHECK_UPDATE = 0.45f;
  static final float PROGRESS_PLAY_SERVICES_UPDATE_FINISHED = 0.7f;
  static final float PROGRESS_WORK_ACCOUNT_REMOVED = 0.8f;
  static final float PROGRESS_ACCOUNT_AUTHENTICATOR_ENABLED = 1.0f;

  @VisibleForTesting
  static final String FEATURE_ARC_DEVICE_MANAGEMENT = "org.chromium.arc.device_management";

  private final Context context;
  private final ComponentName admin;
  private final Handler backgroundHandler;
  private final DevicePolicyManager devicePolicyManager;
  private final DevicePolicyManagerHelper devicePolicyManagerHelper;
  private final WorkAccountApiHelper workAccountApiHelper;
  private final DisableEnrollerAccountSyncHelper disableEnrollerAccountSyncHelper;
  private final DisableSystemInstallers disableSystemInstallers;

  private final boolean supportWorkAccount;
  private WorkingEnvironmentCallback workingEnvironmentCallback;

  private PlayServicesAndStoreUtils playServicesAndStoreUtils;

  EnvironmentPreparer(
      Context context, ComponentName admin, Handler backgroundHandler, boolean supportWorkAccount) {
    this(
        context,
        admin,
        backgroundHandler,
        supportWorkAccount,
        new WorkAccountApiHelper(context),
        new DisableEnrollerAccountSyncHelper(context),
        new DisableSystemInstallers(context, admin));
  }

  @VisibleForTesting
  EnvironmentPreparer(
      Context context,
      ComponentName admin,
      Handler backgroundHandler,
      boolean supportWorkAccount,
      WorkAccountApiHelper workAccountApiHelper,
      DisableEnrollerAccountSyncHelper disableEnrollerAccountSyncHelper,
      DisableSystemInstallers disableSystemInstallers) {
    this.context = context;
    this.admin = admin;
    this.backgroundHandler = backgroundHandler;
    this.devicePolicyManager =
        (DevicePolicyManager) context.getSystemService(Context.DEVICE_POLICY_SERVICE);
    this.devicePolicyManagerHelper = new DevicePolicyManagerHelper(context);
    this.supportWorkAccount = supportWorkAccount;
    this.playServicesAndStoreUtils = new PlayServicesAndStoreUtils(context);
    this.workAccountApiHelper = workAccountApiHelper;
    this.disableEnrollerAccountSyncHelper = disableEnrollerAccountSyncHelper;
    this.disableSystemInstallers = disableSystemInstallers;
  }

  /**
   * Perform all necessary actions to add support for Android for Work Account provisioning.
   *
   * @see AndroidForWorkAccountSupport#ensureWorkingEnvironment
   */
  @WorkerThread
  void prepareEnvironment(final WorkingEnvironmentCallback workingEnvironmentCallback) {
    this.workingEnvironmentCallback = workingEnvironmentCallback;

    if (!playServicesAndStoreUtils.isPlayStoreInstalled()) {
      fail(Error.PLAY_STORE_NOT_FOUND);
      return;
    }

    if (!playServicesAndStoreUtils.isPlayServicesInstalled()) {
      fail(Error.PLAY_SERVICES_NOT_FOUND);
      return;
    }

    if (!useArcPlusPlusDeviceManagement() && !isDeviceProvisioned()) {
      Log.e(Constants.TAG, "SetupWizard is still running. Please wait for "
          + "onProfileProvisioningComplete before calling");
      fail(Error.FAILED_PRECONDITION);
      return;
    }

    PreconditionsHelper preconditionsHelper =
        new PreconditionsHelper(context, admin, devicePolicyManagerHelper);
    if (!preconditionsHelper.ensureHasPermissions(REQUIRED_PERMISSIONS)) {
      Log.e(Constants.TAG, "Must have expected permissions in manifest for provisioning.");
      fail(Error.FAILED_PRECONDITION);
      return;
    }

    if (GoogleApiAvailability.GOOGLE_PLAY_SERVICES_VERSION_CODE
        < PLAY_SERVICES_MIN_VERSION) {
      Log.e(Constants.TAG, "Must have gmscore sdk version at least " + PLAY_SERVICES_MIN_VERSION);
      fail(Error.FAILED_PRECONDITION);
      return;
    }

    // On Android O, creating a work profile breaks WebView on the entire device (personal and work
    // profile). To fix this we enable Chrome on the work profile and prevent the user from
    // disabling it. More context in b/64212702.
    // TODO(faoustin): see if we can check for the bug, since some devices on API 26 will already
    // have the patch.
    if (Build.VERSION.SDK_INT == 26 /* O */ && devicePolicyManagerHelper.isProfileOwner()) {
      try {
        PackageInfo packageInfo =
            context
                .getPackageManager()
                .getPackageInfo(
                    Constants.CHROME_PACKAGE_NAME, PackageManager.MATCH_UNINSTALLED_PACKAGES);
        if ((packageInfo.applicationInfo.flags & ApplicationInfo.FLAG_INSTALLED) == 0) {
          Log.i(Constants.TAG, "Enabling Chrome");
          devicePolicyManager.enableSystemApp(admin, Constants.CHROME_PACKAGE_NAME);
        }
        Log.i(Constants.TAG, "Prevent uninstall of Chrome");
        devicePolicyManager.setUninstallBlocked(admin, Constants.CHROME_PACKAGE_NAME, true);
      } catch (NameNotFoundException e) {
        Log.w(Constants.TAG, "Chrome not found.", e);
      }
    }

    disableEnrollerAccountSyncHelper.disableSyncIfNecessary();

    // For DO and PO provisioning, clean up any blacklisted system installers.
    if (devicePolicyManagerHelper.isProfileOwner() || devicePolicyManagerHelper.isDeviceOwner()) {
      disableSystemInstallers.cleanUpBlacklistedInstallers(Collections.emptyList());
    }

    checkPlayStoreVersion();
  }

  private void success() {
    workingEnvironmentCallback.onSuccess();
  }

  private void fail(final Error error) {
    workingEnvironmentCallback.onFailure(error);
  }

  /**
   * Calls {@code onProgressChange} on {@link #workingEnvironmentCallback} with progress which is
   * expected to be a value between 0 and 1 inclusive.
   */
  private void updateProgress(float progress) {
    workingEnvironmentCallback.onProgressChange(progress);
  }

  /**
   * Checks if the Play Store is sufficiently up-to-date to support Android for Work accounts.
   *
   * <p>Next step:
   *
   * <ul>
   *   <li>if Play Store needs to be updated: {@link #downloadPlayStore},
   *   <li>if Play Store is sufficiently up-to-date: {@link #checkPlayServicesVersion}.
   * </ul>
   */
  private void checkPlayStoreVersion() {
    Log.i(Constants.TAG, "Ensuring Play Store has required version.");
    // To update the Play Store, we must be device owner on an M+ device.
    if (devicePolicyManagerHelper.isDeviceOwner()
        && Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
      if (playServicesAndStoreUtils.isPlayStoreUpToDate()) {
        // Play Store already up-to-date, continue.
        updateProgress(PROGRESS_PLAY_STORE_UPDATE_FINISHED);
        checkPlayServicesVersion();
      } else {
        // Play Store not up-to-date, update it.
        updateProgress(PROGRESS_PLAY_STORE_CHECK_UPDATE);
        downloadPlayStore();
      }
    } else if (playServicesAndStoreUtils.isPlayStoreVersionCompatible()) {
      // Play Store compatible, continue.
      updateProgress(PROGRESS_PLAY_STORE_UPDATE_FINISHED);
      checkPlayServicesVersion();
    } else {
      Log.e(Constants.TAG, "Play Store needs to be updated, but cannot update.");
      fail(Error.PLAY_STORE_OUTDATED);
    }
  }

  /**
   * Downloads the new Play Store APK.
   *
   * <p>Next step: {@link #installPlayStore}.
   */
  private void downloadPlayStore() {
    Log.i(Constants.TAG, "Downloading Play Store.");
    new PlayStoreDownloader(context, backgroundHandler)
        .downloadPlayStore(
            new DownloadCompleteListener() {
              @Override
              public void onComplete(InputStream inputStream) {
                Log.i(Constants.TAG, "Play Store download complete.");
                updateProgress(PROGRESS_PLAY_STORE_DOWNLOADED);
                installPlayStore(inputStream);
              }

              @Override
              public void onFailure(Error reason) {
                Log.i(Constants.TAG, "Play Store download failed.");
                if (playServicesAndStoreUtils.isPlayStoreVersionCompatible()) {
                  // Play Store compatible, continue.
                  updateProgress(PROGRESS_PLAY_STORE_UPDATE_FINISHED);
                  checkPlayServicesVersion();
                } else {
                  fail(reason);
                }
              }

              @Override
              public void onProgressChange(float progress) {
                // Get the step size of progress for current step.
                float stepSize = PROGRESS_PLAY_STORE_DOWNLOADED - PROGRESS_PLAY_STORE_CHECK_UPDATE;
                updateProgress(PROGRESS_PLAY_STORE_CHECK_UPDATE + progress * stepSize);
              }
            });
  }

  /**
   * Installs the Play Store APK. This is only possible when the app is the Device Owner.
   *
   * <p>Next step: {@link #checkPlayServicesVersion}.
   */
  private void installPlayStore(InputStream inputStream) {
    Log.i(Constants.TAG, "Installing Play Store.");
    new PlayStoreInstaller(context, admin, backgroundHandler)
        .installPlayStore(
            inputStream,
            new WorkingEnvironmentCallback() {
              @Override
              public void onSuccess() {
                Log.i(Constants.TAG, "Play Store installation complete.");
                updateProgress(PROGRESS_PLAY_STORE_UPDATE_FINISHED);
                checkPlayServicesVersion();
              }

              @Override
              public void onFailure(Error statusCode) {
                Log.i(Constants.TAG, "Play Store install failed.");
                if (playServicesAndStoreUtils.isPlayStoreVersionCompatible()) {
                  // Play Store compatible, continue.
                  updateProgress(PROGRESS_PLAY_STORE_UPDATE_FINISHED);
                  checkPlayServicesVersion();
                } else {
                  fail(statusCode);
                }
              }
            });
  }

  /**
   * Checks if Play Services is sufficiently up-to-date to support Android for Work accounts.
   *
   * <p>Next step:
   *
   * <ul>
   *   <li>if Play Services needs to be updated: {@link #updatePlayServices},
   *   <li>if Play Services is sufficiently up-to-date and {@link #supportWorkAccount} is true:
   *       {@link #removeEnrollerAccounts},
   *   <li>if Play Services is sufficiently up-to-date and {@link #supportWorkAccount} is false this
   *       is the last step of the flow.
   * </ul>
   */
  private void checkPlayServicesVersion() {
    Log.i(Constants.TAG, "Ensuring Play Services has required version.");
    // We need to be either device or profile owner to update Play Services.
    if (devicePolicyManagerHelper.isDeviceOrProfileOwner()) {
      if (playServicesAndStoreUtils.isPlayServicesUpToDate()) {
        // Play Services already up-to-date, continue.
        updateProgress(PROGRESS_PLAY_SERVICES_UPDATE_FINISHED);
        removeEnrollerAccounts();
      } else {
        // Play Services not up-to-date, update it.
        updateProgress(PROGRESS_PLAY_SERVICES_CHECK_UPDATE);
        updatePlayServices();
      }
    } else if (playServicesAndStoreUtils.isPlayServicesVersionCompatible()) {
      // Play Services compatible, continue.
      updateProgress(PROGRESS_PLAY_SERVICES_UPDATE_FINISHED);
      removeEnrollerAccounts();
    } else {
      Log.e(Constants.TAG, "Play Services needs to be updated, but cannot update.");
      fail(Error.PLAY_SERVICES_OUTDATED);
    }
  }

  /**
   * Updates Play Services by calling Play Store. This is only possible when the app is the Device
   * or Profile owner.
   *
   * <p>Next step:
   *   - if {@link #supportWorkAccount} is true:{@link #removeEnrollerAccounts},
   *   - if {@link #supportWorkAccount} is false this is the last step of the flow.
   */
  private void updatePlayServices() {
    final WorkingEnvironmentCallback updateCompleteCallback =
        new WorkingEnvironmentCallback() {
          @Override
          public void onSuccess() {
            updateProgress(PROGRESS_PLAY_SERVICES_UPDATE_FINISHED);
            removeEnrollerAccounts();
          }

          @Override
          public void onFailure(Error statusCode) {
            if (playServicesAndStoreUtils.isPlayServicesVersionCompatible()) {
              // Play Services compatible, continue.
              updateProgress(PROGRESS_PLAY_SERVICES_UPDATE_FINISHED);
              removeEnrollerAccounts();
            } else {
              fail(statusCode);
            }
          }

          @Override
          public void onProgressChange(float progress) {
            float stepSize =
                PROGRESS_PLAY_SERVICES_UPDATE_FINISHED - PROGRESS_PLAY_SERVICES_CHECK_UPDATE;
            updateProgress(PROGRESS_PLAY_SERVICES_CHECK_UPDATE + progress * stepSize);
          }
        };

    // Ensures the Android checkin has completed before updating Play Services, waiting for it to
    // complete if necessary.
    new CheckinHelper(context, backgroundHandler)
        .ensureCheckinComplete(
            new WorkingEnvironmentCallback() {
              @Override
              public void onSuccess() {
                killPlayIfNecessary();
                Log.i(Constants.TAG, "Updating Play Services.");
                new PlayServicesUpdateScheduler(context, backgroundHandler)
                    .updatePlayServices(updateCompleteCallback);
              }

              @Override
              public void onFailure(Error statusCode) {
                if (playServicesAndStoreUtils.isPlayServicesVersionCompatible()) {
                  // Play Services compatible, continue.
                  updateProgress(PROGRESS_PLAY_SERVICES_UPDATE_FINISHED);
                  removeEnrollerAccounts();
                } else {
                  fail(statusCode);
                }
              }
            });
  }

  /**
   * Trying to kill play in order to workaround an issue that Play cached the empty android id.
   * Note that it is not necessary to apply this in DO case, as android id should be available
   * before Play get started in this case.
   * TODO(b/70147591): Consider to apply the workaround only if play version is < 80841900.
   */
  private void killPlayIfNecessary() {
    if (!devicePolicyManagerHelper.isProfileOwner()) {
      return;
    }
    if (context.checkCallingOrSelfPermission(Manifest.permission.KILL_BACKGROUND_PROCESSES)
            == PackageManager.PERMISSION_GRANTED) {
      Log.i(Constants.TAG, "Trying to kill Play app using killBackgroundProcesses.");
      ActivityManager activityManager =
              (ActivityManager) context.getSystemService(Context.ACTIVITY_SERVICE);
      activityManager.killBackgroundProcesses(Constants.PLAY_STORE_PACKAGE_NAME);
    } else {
      Log.i(Constants.TAG, "Missing KILL_BACKGROUND_PROCESSES, "
            + "use setApplicationHidden to kill Play");
      devicePolicyManager.setApplicationHidden(admin, Constants.PLAY_STORE_PACKAGE_NAME, true);
      devicePolicyManager.setApplicationHidden(admin, Constants.PLAY_STORE_PACKAGE_NAME, false);
    }
  }

  /**
   * Removes the enroller account if there is any. This only applies to Device Owner provisioning.
   *
   * <p>Next step: {@link #enableWorkAccountAuthenticator}.
   */
  private void removeEnrollerAccounts() {
    // Only Android for Work DO mode has enroller accounts, which can be removed.
    if (!devicePolicyManagerHelper.isDeviceOwner()) {
      updateProgress(PROGRESS_WORK_ACCOUNT_REMOVED);
      enableWorkAccountAuthenticator();
      return;
    }

    Log.i(Constants.TAG, "Removing enroller accounts.");
    boolean result = new EnrollerAccountRemover(context).removeEnrollerAccounts();

    if (result) {
      updateProgress(PROGRESS_WORK_ACCOUNT_REMOVED);
      enableWorkAccountAuthenticator();
    } else {
      fail(Error.FAILED_TO_REMOVE_ENROLLER_ACCOUNT);
    }
  }

  /**
   * Enables the authenticator for Android for Work accounts by calling the Play Services work
   * account api.
   *
   * <p>This is the last step of the flow.
   */
  private void enableWorkAccountAuthenticator() {
    if (!supportWorkAccount) {
      // The flow ends here if work account support isn't required.
      success();
      return;
    }

    Log.i(Constants.TAG, "Enabling work account authenticator.");
    Error statusCode =
        new WorkAccountAuthenticatorEnabler(
                context, admin, devicePolicyManagerHelper, workAccountApiHelper)
            .enableWorkAuthenticator();

    if (statusCode == null) {
      // Success.
      updateProgress(PROGRESS_ACCOUNT_AUTHENTICATOR_ENABLED);
      success();
    } else {
      fail(statusCode);
    }
  }

  private boolean useArcPlusPlusDeviceManagement() {
    return context.getPackageManager().hasSystemFeature(FEATURE_ARC_DEVICE_MANAGEMENT);
  }

  private boolean isDeviceProvisioned() {
    return Settings.Global.getInt(context.getContentResolver(),
        Settings.Global.DEVICE_PROVISIONED, 0) == 1;
  }
}
